/*
 *
 *
 *    muxtcpsocket.cc
 *
 *    C++ source code for class my::multiplexed::tcpsocket member functions
 *    and/or static member data.
 *
 *    =========================================================================
 *
 *    Copyright (C) 1999-2003 by Matt Roberts, KK5JY.
 *
 *    This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 *    This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *    =========================================================================
 *
 *
 */


#include <my++/muxtcpsocket.h>
#include "bufcpy.h"

#define DEFAULT_BUFFER_LENGTH (1024)

my::multiplexed::tcpsocket::tcpsocket (void)
	: buffer_len(0), c_buffer(0), is_closed(true),
	is_inbound(false), is_listening(false), tx_ok(false) {
	buffer(DEFAULT_BUFFER_LENGTH);
}

my::multiplexed::tcpsocket::tcpsocket (const inet_address &a, int p)
	: my::tcpsocket(a,p), buffer_len(0), c_buffer(0), is_closed(false),
	is_inbound(false), is_listening(true), tx_ok(false) {
	buffer(DEFAULT_BUFFER_LENGTH);
	is_closed = false;
}

my::multiplexed::tcpsocket::tcpsocket (const address &a)
	: my::tcpsocket(a), buffer_len(0), c_buffer(0), is_closed(false),
	is_inbound(false), is_listening(true), tx_ok(false) {
	buffer(DEFAULT_BUFFER_LENGTH);
}

my::multiplexed::tcpsocket::tcpsocket (const std::string &a, int p)
	: my::tcpsocket(a,p), buffer_len(0), c_buffer(0), is_closed(false),
	is_inbound(false), is_listening(true), tx_ok(false) {
	buffer(DEFAULT_BUFFER_LENGTH);
}

void my::multiplexed::tcpsocket::buffer (size_t len) {
	if (is_closed) {
		if (len > 0) {
			if (c_buffer != 0)
				delete [] c_buffer;
			buffer_len = len;
			c_buffer = new unsigned char [buffer_len];
		}
	}
}


bool my::multiplexed::tcpsocket::should_watch_read(void) {
	return (is_listening || (!is_closed));
}

bool my::multiplexed::tcpsocket::should_watch_write(void) {
	if (is_listening)
		return false;
	else
		return ( ! tx_ok );
}

bool my::multiplexed::tcpsocket::read_action (void) {
	if (is_listening) {
		is_inbound = true;
		return true;
	} else {
		ssize_t rtn = -1;
		do {
			try {
				rtn = my::tcpsocket::get (c_buffer, buffer_len);
			} catch (const tcpsocket::exception &e) {
				if (e.errnum() != EINTR)
					throw;
				rtn = -1;
			}
		} while (rtn < 0);

		if (rtn > 0)
			bufcpy (rx_buffer, c_buffer, rtn);
		else
			is_closed = true;
	}

	return true;
}

bool my::multiplexed::tcpsocket::write_action (void) {
	if (tx_buffer.size() > 0) {
		tx_ok = false;
		size_t len = tx_buffer.size();
		ssize_t rtn = -1;
		if (len > buffer_len)
			len = buffer_len;
		bufcpy (c_buffer, tx_buffer, len);
		do {
			try {
				rtn = my::tcpsocket::put(c_buffer, len);
			} catch (const tcpsocket::exception &e) {
				if (e.errnum() != EINTR)
					throw;
				rtn = -1;
			}
		} while (rtn < 0);
	} else {
		tx_ok = true;
	}

	// no need to bother the application, we took care of it...
	return false;
}

void my::multiplexed::tcpsocket::listen (int backlog) {
	is_listening = true;
	my::tcpsocket::listen (backlog);
}

void my::multiplexed::tcpsocket::listen (void) {
	is_listening = true;
	my::tcpsocket::listen();
}

void my::multiplexed::tcpsocket::accept (socket &s, std::string &a) {
	my::multiplexed::tcpsocket &s2 = dynamic_cast<my::multiplexed::tcpsocket&>(s);
	is_inbound = false;
	my::tcpsocket::accept(s2, a);
	s2.is_closed = false;
	s2.is_listening = false;
}

void my::multiplexed::tcpsocket::accept (socket &s, socket::address &a) {
	my::multiplexed::tcpsocket &s2 = dynamic_cast<my::multiplexed::tcpsocket&>(s);
	is_inbound = false;
	my::tcpsocket::accept(s2, a);
	s2.is_closed = false;
	s2.is_listening = false;
}

void my::multiplexed::tcpsocket::accept (socket &s) {
	my::multiplexed::tcpsocket &s2 = dynamic_cast<my::multiplexed::tcpsocket&>(s);
	is_inbound = false;
	my::tcpsocket::accept(s2);
	s2.is_closed = false;
	s2.is_listening = false;
}

void my::multiplexed::tcpsocket::connect (const socket::address &a) {
	my::tcpsocket::connect (a);
	is_closed = false;
	is_listening = false;
}

void my::multiplexed::tcpsocket::connect (const std::string &a) {
	my::tcpsocket::connect (a);
	is_closed = false;
	is_listening = false;
}

size_t my::multiplexed::tcpsocket::get (unsigned char *buf, size_t len) {
	size_t rtn = rx_buffer.size();
	if (len < rtn)
		rtn = len;
	if (rtn > 0)
		bufcpy (buf, rx_buffer, rtn);
	return rtn;
}

size_t my::multiplexed::tcpsocket::get (char *buf, size_t len) {
	size_t rtn = rx_buffer.size();
	if (len < rtn)
		rtn = len;
	if (rtn > 0)
		bufcpy ((unsigned char*)buf, rx_buffer, rtn);
	return rtn;
}

size_t my::multiplexed::tcpsocket::get (void *buf, size_t len) {
	size_t rtn = rx_buffer.size();
	if (len < rtn)
		rtn = len;
	if (rtn > 0)
		bufcpy (static_cast<unsigned char *>(buf), rx_buffer, rtn);
	return rtn;
}

size_t my::multiplexed::tcpsocket::get (std::string &s) {
	size_t rtn = rx_buffer.size();
	bufcpy(s, rx_buffer);
	return rtn;
}

size_t my::multiplexed::tcpsocket::put (const std::string &s) {
	if (!is_closed) {
		if (tx_ok) {
			tx_ok = false;
			size_t rtn = my::tcpsocket::put(s.c_str(), s.size());
			if (rtn != s.size()) {
				std::string remainder(s.begin() + rtn, s.end());
				bufcpy(tx_buffer, remainder);
			}
		} else {
			bufcpy(tx_buffer, s);
		}
	} else {
		return 0;
	}

	return s.size();
}

size_t my::multiplexed::tcpsocket::put (const unsigned char *buf, size_t len) {
	if (!is_closed) {
		if (tx_ok) {
			tx_ok = false;
			size_t rtn = my::tcpsocket::put(buf, len);
			if (rtn != len)
				bufcpy (tx_buffer, buf + rtn, len - rtn);
		} else {
			bufcpy (tx_buffer, buf, len);
		}
	} else {
		return 0;
	}

	return len;
}

size_t my::multiplexed::tcpsocket::put (const char *buf, size_t len) {
	if (!is_closed) {
		if (tx_ok) {
			tx_ok = false;
			size_t rtn = my::tcpsocket::put(buf, len);
			if (rtn != len)
				bufcpy (tx_buffer, (const unsigned char*)(buf + rtn), len - rtn);
		} else {
			bufcpy (tx_buffer, (const unsigned char*)(buf), len);
		}
	} else {
		return 0;
	}

	return len;
}

size_t my::multiplexed::tcpsocket::put (const void *buf, size_t len) {
	if (!is_closed) {
		if (tx_ok) {
			tx_ok = false;
			size_t rtn = my::tcpsocket::put(buf, len);
			if (rtn != len)
				bufcpy (tx_buffer, static_cast<const unsigned char *>(buf) + rtn, len - rtn);
		} else {
			bufcpy (tx_buffer, static_cast<const unsigned char *>(buf), len);
		}
	} else {
		return 0;
	}

	return len;
}
